#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gtk/gtk.h>

typedef struct png_plot_t png_plot;

struct png_plot_t {
    GtkWidget *shell;
    GtkWidget *canvas;
    GtkWidget *statusarea;    
    GtkWidget *cursor_label;
    GdkPixmap *pixmap;
    double xmin, xmax;
    double ymin, ymax;
    int pixel_width, pixel_height;
    int pixel_xmin, pixel_xmax;
    int pixel_ymin, pixel_ymax;
    int xint, yint;
};

static void errbox (const char *s)
{
    fprintf(stderr, "%s\n", s);
}

static png_plot *png_plot_new (void)
{
    png_plot *plot = malloc(sizeof *plot);

    if (plot == NULL) {
	return NULL;
    }

    plot->shell = NULL;
    plot->canvas = NULL;
    plot->statusarea = NULL;    
    plot->cursor_label = NULL;
    plot->pixmap = NULL;

    plot->pixel_width = 680; /* 640 */
    plot->pixel_height = 400; /* 480 */

    plot->xmin = plot->xmax = 0.0;
    plot->ymin = plot->ymax = 0.0;
    plot->xint = plot->yint = 0;

    return plot;
}

static void destroy_png_plot (GtkWidget *w, png_plot *plot)
{
    gtk_widget_unref(plot->shell);
    free(plot);
    gtk_main_quit();
}

static 
void plot_expose (GtkWidget *widget, GdkEventExpose *event,
		  GdkPixmap *dbuf_pixmap)
{
    /* Don't repaint entire window on each exposure */
    gdk_window_set_back_pixmap(widget->window, NULL, FALSE);

    /* Refresh double buffer, then copy the "dirtied" area to
       the on-screen GdkWindow */
    gdk_window_copy_area(widget->window,
			 widget->style->fg_gc[GTK_STATE_NORMAL],
			 event->area.x, event->area.y,
			 dbuf_pixmap,
			 event->area.x, event->area.y,
			 event->area.width, event->area.height);
}

#define float_fmt(i,x) ((i) && fabs(x) < 1.0e7)

static gint
motion_notify_event (GtkWidget *widget, GdkEventMotion *event, png_plot *plot)
{
    GdkModifierType state;
    gchar label[32], label_y[16];
    int x, y;

    if (event->is_hint) {
        gdk_window_get_pointer(event->window, &x, &y, &state);
    } else {
        x = event->x;
        y = event->y;
        state = event->state;
    }

    *label = 0;

    if (x > plot->pixel_xmin && x < plot->pixel_xmax && 
	y > plot->pixel_ymin && y < plot->pixel_ymax) {
	double xspan = plot->pixel_xmax - plot->pixel_xmin;
	double yspan = plot->pixel_ymax - plot->pixel_ymin;
	double data_x, data_y;

	data_x = plot->xmin + ((double) x - plot->pixel_xmin) / 
	    xspan * (plot->xmax - plot->xmin);
	data_y = plot->ymax - ((double) y - plot->pixel_ymin) / 
	    yspan * (plot->ymax - plot->ymin);
	sprintf(label, (float_fmt(plot->xint, data_x))? "%7.0f" : 
		"%7.4g", data_x);
	sprintf(label_y, (float_fmt(plot->yint, data_y))? " %-7.0f" : 
		" %-7.4g", data_y);
	strcat(label, label_y);
    } 

    gtk_label_set_text(GTK_LABEL(plot->cursor_label), label);
  
    return TRUE;
}

int render_pngfile (png_plot *plot, const char *fname)
{
    gint width, height;
    GdkPixbuf *pbuf;
    GError *error = NULL;

    pbuf = gdk_pixbuf_new_from_file(fname, &error);
    if (pbuf == NULL) {
        errbox(error->message);
        g_error_free(error);
	return 1;
    }

    width = gdk_pixbuf_get_width(pbuf);
    height = gdk_pixbuf_get_height(pbuf);

    if (width == 0 || height == 0) {
	errbox("Malformed PNG file for graph");
	g_object_unref(pbuf);
	return 1;
    }

    gdk_pixbuf_render_to_drawable(pbuf, plot->pixmap, 
				  plot->canvas->style->fg_gc[GTK_STATE_NORMAL],
				  0, 0, 0, 0, width, height,
				  GDK_RGB_DITHER_NONE, 0, 0);

    g_object_unref(pbuf);
    
    return 0;
}

static int get_plot_ranges (png_plot *plot, const char *dimfile)
{
    FILE *fp;
    char line[128];
    int ybot, ytop;
    int got_bounds = 0;
    int err = 0;

    fp = fopen(dimfile, "r");
    if (fp == NULL) {
	return 1;
    }

    while (fgets(line, sizeof line, fp)) {
	if (sscanf(line, "pixel_bounds: %d %d %d %d",
		   &plot->pixel_xmin, &plot->pixel_xmax,
		   &ybot, &ytop) == 4) {
	    got_bounds++;
	} else if (sscanf(line, "data_bounds: %lf %lf %lf %lf", 
			  &plot->xmin, &plot->xmax,
			  &plot->ymin, &plot->ymax) == 4) {
	    got_bounds++;
	}
    }

    fclose(fp);

    if (got_bounds != 2) {
	fprintf(stderr, "get_png_bounds_info(): failed\n");
	return 1;
    }

    plot->pixel_ymin = plot->pixel_height - ytop;
    plot->pixel_ymax = plot->pixel_height - ybot;

    if ((plot->xmax - plot->xmin) / 
	(plot->pixel_xmax - plot->pixel_xmin) >= 1.0) {
	plot->xint = 1;
    }
    if ((plot->ymax - plot->ymin) / 
	(plot->pixel_ymax - plot->pixel_ymin) >= 1.0) {
	plot->yint = 1;
    }

    return err;
}

static png_plot *show_gnuplot_png (const char *pngfile,
				   const char *dimfile)
{
    GtkWidget *vbox;
    GtkWidget *canvas_hbox;
    GtkWidget *label_frame = NULL;
    GtkWidget *status_hbox = NULL;
    png_plot *plot;
    int err = 0;

    plot = png_plot_new();
    if (plot == NULL) {
	return NULL;
    }  

    err = get_plot_ranges(plot, dimfile);
    if (err) {
	free(plot);
	return NULL;
    }

    plot->shell = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_ref(plot->shell);

    gtk_window_set_title(GTK_WINDOW(plot->shell), "gnuplot PNG graph"); 
    gtk_window_set_resizable(GTK_WINDOW(plot->shell), FALSE);

    vbox = gtk_vbox_new(FALSE, 2);
    gtk_container_add(GTK_CONTAINER(plot->shell), vbox);

    g_signal_connect(G_OBJECT(plot->shell), "destroy",
		     G_CALLBACK(destroy_png_plot), plot);

    /* box to hold canvas */
    canvas_hbox = gtk_hbox_new(FALSE, 1);
    gtk_box_pack_start(GTK_BOX(vbox), canvas_hbox, TRUE, TRUE, 0);
    gtk_widget_show(canvas_hbox);

    /* eventbox and hbox for status area  */
    plot->statusarea = gtk_event_box_new();
    gtk_box_pack_start(GTK_BOX(vbox), plot->statusarea, FALSE, FALSE, 0);

    status_hbox = gtk_hbox_new (FALSE, 2);
    gtk_container_add(GTK_CONTAINER(plot->statusarea), status_hbox);
    gtk_widget_show (status_hbox);
    gtk_container_set_resize_mode(GTK_CONTAINER (status_hbox),
				  GTK_RESIZE_QUEUE);

    /* Create drawing-area widget */
    plot->canvas = gtk_drawing_area_new();
    gtk_widget_set_size_request(GTK_WIDGET(plot->canvas), 
				plot->pixel_width, plot->pixel_height);
    gtk_widget_set_events (plot->canvas, GDK_EXPOSURE_MASK
                           | GDK_LEAVE_NOTIFY_MASK
                           | GDK_POINTER_MOTION_MASK
                           | GDK_POINTER_MOTION_HINT_MASK);

    GTK_WIDGET_SET_FLAGS(plot->canvas, GTK_CAN_FOCUS);

    /* cursor label (graph position indicator) */
    label_frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(label_frame), GTK_SHADOW_IN);
    plot->cursor_label = gtk_label_new(" ");
    gtk_container_add(GTK_CONTAINER(label_frame), plot->cursor_label);
    gtk_widget_show(plot->cursor_label);

    g_signal_connect(G_OBJECT(plot->canvas), "motion_notify_event",
		     G_CALLBACK(motion_notify_event), plot);

    /* pack the widgets */
    gtk_box_pack_start(GTK_BOX(canvas_hbox), plot->canvas, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(status_hbox), label_frame, FALSE, FALSE, 0);
    gtk_widget_show(label_frame);

    /* show stuff */
    gtk_widget_show(plot->canvas);
    gtk_widget_show(plot->statusarea);

    gtk_widget_realize(plot->canvas);
    gdk_window_set_back_pixmap(plot->canvas->window, NULL, FALSE);

    gtk_widget_realize(plot->cursor_label);
    gtk_widget_set_size_request(plot->cursor_label, 140, -1);

    gtk_widget_show(vbox);
    gtk_widget_show(plot->shell);       

    /* set the focus to the canvas area */
    gtk_widget_grab_focus(plot->canvas);  

    plot->pixmap = gdk_pixmap_new(plot->shell->window, 
				  plot->pixel_width, plot->pixel_height, 
				  -1);
    g_signal_connect(G_OBJECT(plot->canvas), "expose_event",
		     G_CALLBACK(plot_expose), plot->pixmap);

    err = render_pngfile(plot, pngfile);

    if (err) {
	gtk_widget_destroy(plot->shell);
	plot = NULL;
    } else {
	gtk_main();
    }

    return plot;
}

int main (int argc, char **argv)
{
    struct png_plot_t *plot;
    char *pngfile;
    char *dimfile;

    if (argc != 3) {
	fprintf(stderr, "%s: please give the name of a PNG file and "
		"a bounds file\n", argv[0]);
	exit(EXIT_FAILURE);
    }

    gtk_init(&argc, &argv);

    pngfile = argv[1];
    dimfile = argv[2];

    plot = show_gnuplot_png(pngfile, dimfile);

    return 0;
}

